# Copyright: Peter Norvig, 2002.
import operator
from norvig_utils import *

class Node:
    """A node in a search tree. Contains a pointer to the parent (the node
    that this is a successor of) and to the actual state for this node. Note
    that if a state is arrived at by two paths, then there are two nodes with
    the same state.  Also includes the action that got us to this state, and
    the total path_cost (also known as g) to reach the node.  Other functions
    may add an f and h value; see best_first_graph_search and astar_search for
    an explanation of how the f and h values are handled. You will not need to
    subclass this class."""

    def __init__(self, state, parent=None, action=None, path_cost=0):
        "Create a search treee Node, derived from a parent by an action."
        self.state, self.parent, self.action, self.path_cost, = (
             state,      parent,      action,      path_cost)
        if parent:
            self.depth = parent.depth + 1
        else:
            self.depth = 0

    def __repr__(self):
        return "<Node %s>" % self.state

    def path(self):
        "Create a list of nodes from the root to this node."
        x, result = self, [self]
        while x.parent:
            result.append(x.parent)
            x = x.parent
        return result

    def expand(self, problem):
        "Return a list of nodes reachable from this node. [Fig. 3.8]"
        return [Node(next, self, act,
                     problem.path_cost(self.path_cost, self.state, act, next))
                for (act, next) in problem.successor(self.state)]

class Problem:
    """The abstract class for a formal problem.  You should subclass this and
    implement the method successor, and possibly __init__, goal_test, and
    path_cost. Then you will create instances of your subclass and solve them
    with the various search functions."""

    def __init__(self, initial, goal=None):
        """The constructor specifies the initial state, and possibly a goal
        state, if there is a unique goal.  Your subclass's constructor can add
        other arguments."""
        self.initial = initial; self.goal = goal

    def successor(self, state):
        """Given a state, return a sequence of (action, state) pairs reachable
        from this state. If there are many successors, consider an iterator
        that yields the successors one at a time, rather than building them
        all at once. Iterators will work fine within the framework."""
        abstract()

    def goal_test(self, state):
        """Return True if the state is a goal. The default method compares the
        state to self.goal, as specified in the constructor. Implement this
        method if checking against a single self.goal is not enough."""
        return state == self.goal

    def path_cost(self, c, state1, action, state2):
        """Return the cost of a solution path that arrives at state2 from
        state1 via action, assuming cost c to get up to state1. If the problem
        is such that the path doesn't matter, this function will only look at
        state2.  If the path does matter, it will consider c and maybe state1
        and action. The default method costs 1 for every step in the path."""
        return c + 1


class OurProblem(Problem):
    def __init__(self, initial, goal, map):
        Problem.__init__(self, initial, goal)
        self.map = map

    def successor(self, state):
        x, y = state
        poss = []
        for dx, dy in ((0,1),(0,-1),(1,0),(-1,0)):
            if self.map[y+dy][x+dx] == '.':
                poss.append((x+dx,y+dy))
        return poss

class PriorityQueue:
    "A queue in which the minimum element (as determined by lt) is first."
    def __init__(self, lt=operator.lt): self.A = []; self.lt = lt

    def extend(self, l):
        for i in l: self.append(l)

    def append(self, item):
        A = self.A
        A.append(item)
        i = len(A) - 1
        # Parent is i/2 # In Python2.2, this will be i//2
        while i > 0 and self.lt(item, A[i/2]):
            A[i], i = A[i/2], i/2
        A[i] = item

    def __len__(self): return len(self.A)

    def pop(self):
        if len(self.A) == 1: return self.A.pop()
        e = self.A[0]
        self.A[0] = self.A.pop()
        self.heapify(0)
        return e

    def heapify(self, i):
        "Assumes A is an array such that left(i) and right(i) are heaps,"
        "move A[i] into the correct position.  See CLR&S p. 130"
        A, lt = self.A, self.lt
        left, right, N = 2*i + 1, 2*i + 2, len(A)-1
        if left <= N and lt(A[left], A[i]):
            smallest = left
        else:
            smallest = i
        if right <= N and lt(A[right], A[smallest]):
            smallest = right
        if smallest != i:
            A[i], A[smallest] = A[smallest], A[i]
            self.heapify(smallest)


def graph_search(problem, fringe):
    """Search through the successors of a problem to find a goal.
    The argument fringe should be an empty queue.
    If two paths reach a state, only use the best one. [Fig. 3.18]"""
    closed = {}
    fringe.append(Node(problem.initial))
    while fringe:
        node = fringe.pop()
        if problem.goal_test(node.state):
            return node
        if node.state not in closed:
            closed[node.state] = True
            fringe.extend(node.expand(problem))
    return None


def best_first_graph_search(problem, f):
    """Search the nodes with the lowest f scores first.
    You specify the function f(node) that you want to minimize; for example,
    if f is a heuristic estimate to the goal, then we have greedy best
    first search; if f is node.depth then we have depth-first search.
    There is a subtlety: the line "f = memoize(f, 'f')" means that the f
    values will be cached on the nodes as they are computed. So after doing
    a best first search you can examine the f values of the path returned."""
    f = memoize(f, 'f')
    return graph_search(problem, PriorityQueue(lambda a,b: f(a) < f(b)))


def astar_search(problem, h=None):
    """A* search is best-first graph search with f(n) = g(n)+h(n).
    You need to specify the h function when you call astar_search.
    Uses the pathmax trick: f(n) = max(f(n), g(n)+h(n))."""
    h = h or problem.h
    def f(n):
        print n
        return max(getattr(n, 'f', -infinity), n.path_cost + h(n))
    return best_first_graph_search(problem, f)


def try_with_icfp_map(map, start, end, h):
    node = astar_search(OurProblem(start, end, map), h)
    path = node.path()
    return [n.state for n in path]

