#!/usr/bin/env python

"""
An interactively debuggable bot base class.
To use in a separate file:
class MyBot(Debugot.Debugot):
   def think(self): return 'reasoning', 'command'

if __name__ == '__main__':
   Debugot.main(Mybot)

The program can be run with none, 2 of 3 args:
    none: localhost, 4000, .25 s timeout
    2: #1 is host, #2 is port
    3: #3 is timeout in seconds (float)
"""
from twisted.internet import default; default.install()
from twisted.internet import reactor
import sys, random, select, operator, string, os, time
import icfp
from best_path import try_with_icfp_map
import astar

def showMap(m, top=0,left=0,width=0,height=0):
    "generate and return a possibly partial map"
    left = max(0, left)
    top = max(0,top)
    if width == 0:
        width = len(m[0]) - left
    if height == 0:
        height = len(m) - top

    width = min(width, len(m[0]) - left)
    height = min(height, len(m) - top)
    r = ""
    for pos in xrange(3):
        r += "    %s\n" % ''.join([('%3d' % x)[pos] for x in range(left, left+width)])
    for x in xrange(top, top+height):
        r += "%3d %s\n" % (x, m[x][left:left+width])

    return r

def manhattan(a,b):
    return abs(a[0] - b[0]) + abs(a[1] - b[1])

RCOORDS = {}                            # Reverse coordinates
for k,v in icfp.COORDS.items():
    RCOORDS[v] = k

LOG_SIZE = 15

ROBOT_NAMES = string.letters

def makeName(id):
    return ROBOT_NAMES[id % len(ROBOT_NAMES)]

BOLD = "\033[31;1m"
BLUE = "\033[34;1m"
GREEN = "\033[32;1m"
OFF = "\033[0m"
CLEAR_SCREEN = "\033[2J\033[1;1H"
LOG_WIDTH = 40
class Debugot(icfp.ICFP2002ClientProtocol):

    def __init__(self, interactive=2, timeout=None):
        self.userTimeout = timeout      # Timeout for user input
        self.log = []                   # Cyclic robot log
        self.turn = 0                   # Current turn
        self.columns = int(os.environ.get('COLUMNS', 80))
        self.rows = int(os.environ.get('ROWS', 80))
        self.lastPacket = time.time()
        self.totalWait = 0
        if interactive >= 3:
            self.endOfTurn = self.interactiveEndOfTurn
        elif interactive >= 1:
            self.endOfTurn = self.simpleLoggingEndOfTurn
            self.addLog = self.printLog
        elif interactive == 0:
            self.endOfTurn = self.simpleLoggingEndOfTurn
            self.addLog = lambda x: None
            
        self.interactive = interactive

    def addLog(self, entry):
        self.log.append('%3d  %s' % (self.turn, entry[:LOG_WIDTH]))
        self.log = self.log[-LOG_SIZE:]

    def printLog(self, entry):
        print '%3d: %s' % (self.turn, entry)

    def connectionMade(self):
        print 'Connection made'
        icfp.ICFP2002ClientProtocol.connectionMade(self)
    
    def think(self):
        "Override this and return tuple of reasoning, command the robot would execute"
        raise NotImplementedError

    def getKnownPackageCount(self):
        return reduce(operator.add, map(len, self.knownPackages.values()), 0)

    def formatPackages(self, plist,weightLeft=0):
        if len(plist) > 100:
            print " [>100 packages]"
            return
        packages = {}
        for p in plist:
            packages.setdefault((p.weight, p.x, p.y), []).append(p)
        for l in packages.values():
            p = l[0]
            print ' %c%s (%2d, %2d)' % ( ' *'[p.weight <= weightLeft],
                                       p.weight, p.x, p.y),
            if len(l) > 20:
                print " [%d packages] " % len(l)
            else:
                for x in l:
                    print '%2d' % x.id,
                print
        
    
    def interactiveEndOfTurn(self):
        self.turn += 1
        self.totalWait += time.time() - self.lastPacket
        m = self.map[:]
        for r in self.knownRobots.values():
            name = makeName(r.id)
            m[r.y] = m[r.y][:r.x] + name + m[r.y][r.x+1:]

        sys.stdout.write(CLEAR_SCREEN)
        # Show at most 40x20 or so as not to fill up the entire screen

        width = self.columns - (LOG_WIDTH+5)
        height = self.rows - 10         # roughly
        left,top = self.getPosition()
        left -= width/2; top -= height/2

       
        m = showMap(m,top,left,width,height).replace(makeName(self.number), BOLD + makeName(self.number) + OFF).replace(icfp.WATER, BLUE + icfp.WATER + OFF).replace(
            icfp.BASE, GREEN+icfp.BASE + OFF)
        formatted = m.split('\n')

        utime = os.times()[0]
        log = ['        YOUR ID: %d -- %s%s%s    CPU: %3.1fs, %3d ms/t  Ping: %dms' % (
            self.number, BOLD, makeName(self.number), OFF,
            utime, 1000 * (utime/self.turn),
            (1000 * self.totalWait) / self.turn
            ),
               '','Turn: %4d  Points: %4d/%4d  Pkgs: %d  DL: %d' % (
            self.turn, self.points, self.totalPoints, self.getKnownPackageCount(),
            self.deadlockLevel
            ), astar.getStats(), ''] +  self.log[:]  # Some spacer on top
        # Show robot log next to the map
        if len(log) < len(formatted):
            log += [''] * (len(formatted) - len(log))
        else:
            formatted += [' ' * 40] * (len(log) - len(formatted))

        for x,y in zip(formatted, log):
            print "%s %s" % (x,y)

        print 'Robots:'
        for r in self.knownRobots.values():
            print ' #%s: %s (%s, %s) %s' % (r.id, makeName(r.id),
                                            r.x, r.y, ' '.join(map(str,r.packages)))

        print 'My packages:'
        self.formatPackages(self.myPackages)

        weightLeft = self.weight - self.getCarryWeight()
        if self.knownPackages.has_key(self.getPosition()):
            print 'Packages on this square:'
            self.formatPackages(self.knownPackages[self.getPosition()],weightLeft)


        sys.stdout.flush()
        done = 0
        while not done:
            robotReasoning, robotCommand = self.think()

            print "Robot thinks: \n %s" % robotReasoning
            
            print '(%d,%d $%d, %du) Command [%s]: ' % (
                self.getPosition() + (self.money, weightLeft, robotCommand)),
            sys.stdout.flush()

            # Wait with timeout for user input
            readable, _, _ = select.select([sys.stdin], [], [], self.userTimeout)
            if sys.stdin in readable:
                cmd = raw_input()
                if cmd == '':
                    cmd = robotCommand
                    if self.userTimeout:
                        sel.userTimeout = None # Stop timeout
            else:
                print '[wait]',; sys.stdout.flush()
                cmd = robotCommand
            done = self.parseCommand(cmd)
               
        self.lastPacket = time.time()

    def parseCommand(self, cmd):
        # EZ-command
        if cmd[0] in 'nesw':
            self.moveRobot(1, cmd[0].upper())
            return 1

        parts = cmd.split()

        if not parts[0].isdigit():
            # No bid supplied, assuming 1
            parts = ['1'] + parts

        bid = int(parts[0]); done = 1
        name = parts[1].lower()
        if 'move'.startswith(name):
            self.moveRobot(bid, parts[2])
        elif 'pick'.startswith(name):
            self.getPackage(bid, map(int, parts[2:]))
        elif 'drop'.startswith(name):
            self.dropPackage(bid, map(int, parts[2:]))
        elif 'timeout'.startswith(name):
            self.userTimeout = float(parts[2]); done = 0
            print "Timeout set to %f seconds" % self.userTimeout
        elif 'path'.startswith(name):
            coord = map(int, parts[2:4])
            print "Path to %d,%d:" % (coord[0], coord[1]),
            res = self.findpath(coord)
            if res is None:
                print "NOT FOUND"
            else:
                for y,x in res:
                    print "%d,%d" % (x,y),
                print
            done = 0;
        elif name[0] == '!':
            return self.parseCommand(self.robotCommand(name[1:], parts[2:]))
        else:
            print 'Invalid command: %s' % cmd
            done = 0

        return done
        

    def simpleLoggingEndOfTurn(self):
        self.turn += 1
        robotReasoning, robotCommand = self.think()
        if self.interactive >= 2:
            print "\n> %s" % robotCommand,
            sys.stdout.flush()
        while not self.parseCommand(robotCommand):
            pass

    def findpath_norvig(self, coord):
        return try_with_icfp_map(self.map, (self.myRobot.y, self.myRobot.x), coord, manhattan)

    def findpath_astar(self, coord):
        "Use (x,y) for coordinates"
        return astar.astar((self.myRobot.y, self.myRobot.x), (coord[1], coord[0]), self.map)

    findpath = findpath_astar
                

class StupidClient(Debugot):
    def think(self):
        return "I'm just a stupid bot!", random.choice('nesw')

def main(factory):
    # Improve performance by redrawing less
    if len(sys.argv) == 1:
        host,port, timeout = 'localhost', 4000, .25
    else:
        host, port = sys.argv[1:3]
        port = int(port)
        if len(sys.argv) == 4:
            timeout = float(sys.argv[3])
        else:
            timeout = None

    interactive = int(os.environ.get('INTERACTIVE', 2))
    if interactive >= 3:
        sys.stdout = open("/dev/tty","w", 16384)
    else:
        timeout = 0
            
    reactor.clientTCP(host, port, factory(interactive, timeout))
    # XXX I DONT CARE IF YOU ARE DEPRECIATED
    reactor.run()

if __name__ == '__main__':
    main(StupidClient)

